package clsp.aligner;

import java.io.PrintWriter;

import clsp.util.AER;
import clsp.util.Alignment;
import clsp.util.IOUtils;
import clsp.util.Logger;
import clsp.util.Option;
import clsp.util.OptionsParser;
import clsp.util.SentencePair;

/**
 * Entry point for word alignment.
 */
public class Main implements Runnable {

	// Training regimen
	@Option(gloss = "Number of iterations to run the model.")
	public int numIterations = 1;
	@Option(gloss = "Whether to train the two models jointly or independently.")
	public TrainMode trainingMode = TrainMode.FORWARD;
	@Option(gloss = "What type of model to estimate")
	public ModelT modelType = ModelT.DICE;

	// Testing and aligning
	@Option(gloss = "Alignment threshold")
	public double threshold = 0.5;
	@Option(gloss = "Print a number of sample alignments for inspection")
	public int numAlignmentsToPrint = 5;
	@Option(gloss = "Root name of Pharaoh-format files for translation")
	public String trainingAlignmentsFile = null;

	public void run() {
		// Which models to train (1 normal vs. 2 reversed)
		boolean b1 = trainingMode != TrainMode.REVERSE;
		boolean b2 = trainingMode != TrainMode.FORWARD;

		Logger.track("Initializing models");
		Model wa1 = null, wa2 = null;
		if (b1) {
			wa1 = modelType.getModel(false);
			wa1.initialize(data.trainingPairs);
		}
		if (b2) {
			wa2 = modelType.getModel(true);
			if (b2) wa2.initialize(data.trainingPairs);
		}
		Logger.end_track();

		// Train
		for (int iter = 0; iter < numIterations; iter++) {
			Logger.track("Training iteration %d", iter + 1);
			for (SentencePair pair : data.trainingPairs) {
				if (b1) wa1.observe(pair);
				if (b2) wa2.observe(pair);
			}
			if (b1) wa1.completeIteration(iter);
			if (b2) wa2.completeIteration(iter);
			Logger.end_track();
		}

		// A WordAligner chooses how to interpret the soft output of a model
		WordAligner aligner = new ThresholdAligner(wa1, wa2, threshold);

		// Evaluate
		if (data.testPairs.size() > 0) {
			Logger.track("Evaluating");
			AER aer = evaluate(aligner, numAlignmentsToPrint);
			Logger.end_track();
			Logger.logss("Performance: %s", aer.toString());
		}

		// Output aligned training data for Joshua
		if (trainingAlignmentsFile != null) {
			Logger.track("Aligning training set");
			PrintWriter out = IOUtils.openOutHard(trainingAlignmentsFile);
			for (SentencePair pair : data.trainingPairs) {
				Alignment al = aligner.alignSentencePair(pair);
				out.println(al.outputPharaohHard());
			}
			out.close();
			Logger.end_track();
		}
	}

	private AER evaluate(WordAligner aligner, int numToPrint) {
		AER aer = new AER();
		int n = 0;
		for (SentencePair pair : data.testPairs) {
			Alignment al = aligner.alignSentencePair(pair);
			Alignment ref = pair.getAlignment();
			if (n < numToPrint) {
				Logger.logss("---Alignment %d---\n%s", ++n, Alignment.render(ref, al));
			}
			aer.addAlignment(al, ref);
		}
		aer.computeFromCounts();
		return aer;
	}

	public enum ModelT {
		DICE {
			@Override
			public Model getModel(boolean reverse) {
				return new HeuristicModel(reverse);
			}
		},
		MODEL1 {
			@Override
			public Model getModel(boolean reverse) {
				return new Model1(reverse);
			}
		};

		public abstract Model getModel(boolean reverse);
	};

	public enum TrainMode {
		FORWARD, REVERSE, BOTH
	};

	static Data data; // Where all the training/test sentences live

	public static void main(String[] args) {
		Main main = new Main();
		data = new Data();
		OptionsParser parser = new OptionsParser();
		parser.doRegister("logger", Logger.class);
		parser.doRegister("main", main);
		parser.doRegister("data", data);
		parser.doParse(args);
		Logger.init();
		Logger.track("main");
		data.loadData();
		main.run();
		Logger.end_track();
	}
}
